/*
	ArrayManager.c: routines from ArrayManager package
		
	These routines are not public domain, but may be used freely
	for non-commercial purposes.
	
	Copyright 1987-1992, Josef Goettgens. All rights reserved.
*/

#include <math.h>
#include <stdlib.h>
#include "RedMech.h"

#define AM_Fatal(s)	FatalError(s)


void DisposeMatrix( MatrixPtr m )
{
	Free2DArray( m->mat );
	free( m );
}



void Free2DArray( double **p )
{
	free( p );
}


void Free1DArray( double *ptr )
{
	free( ptr );
}



MatrixPtr NewMatrix( int phys_rows, int phys_cols, int partition )
{
	MatrixPtr m = (MatrixPtr)malloc( sizeof(Matrix) );
	if ( m == NULL ) AM_Fatal( "NewMatrix(): memory allocation failed" );
	switch ( partition ) {
		case kRowPointers:
			m->mat = New2DArray( phys_rows, phys_cols );
			break;
		case kColumnPointers:
			m->mat = New2DArray( phys_cols, phys_rows );
			break;
		default:
			AM_Fatal( "NewMatrix(): unknown partition type" );
	}
	
	m->rows = m->phys_rows = phys_rows;
	m->cols = m->phys_cols = phys_cols;
	m->partition = partition;
	
	return m;
}



double *New1DArray( int n )
{
	double	*ptr = NULL;

	ptr = (double *)malloc( sizeof(double) * n );
	if ( !ptr ) AM_Fatal( "New1DArray(): Memory allocation failed" );
	
	ClearArray( ptr, n );
	
	return ptr;
}


void ClearArray( register double *ptr, register int len )
{
	const FPUType zero = 0.0;
	register int i;
	
	for ( i = 0; i < len; ++i ) *ptr++ = zero;
}



double **New2DArray( int rows, int columns )
{
	const int kAlign = 8;
	char	*pool = NULL;
	double	**p = NULL, *a = NULL;
	int		size, offset;
	int		i;

	offset = rows * sizeof( a );
	if ( offset % kAlign ) offset += kAlign - (offset % kAlign);
	size = rows * columns * sizeof( *a ) + offset;
	pool = (char *)malloc( (size_t)size );
	if ( !pool ) AM_Fatal( "New2DArray(): Memory allocation failed" );
	
	p = (double **)pool;
	a = (double *)(pool + offset);
	for ( i = 0; i < rows; ++i ) *(p + i) = a + i * columns;
	ClearArray( a, rows * columns );
	
	return p;
}



void MinMax( const double *x, int n, double *maxi, double *mini )
{
	int		i;
	FPUType	ma, mi;
	
	ma = mi = *x++;
	for ( i = 1; i < n; ++i, ++x ) {
		if ( *x > ma ) ma = *x;
		if ( *x < mi ) mi = *x;
	}
	*maxi = (double)ma;
	*mini = (double)mi;
}


