/*
	Scanner.c

	RedMech is a tool for automatically reducing CnHm-mechanisms.
	Written by J. Goettgens, P. Terhoeven, R. Korff	
*/

#ifdef applec
#pragma load "RedMech.dump"
#else
#include "RedMechHeaders.h"
#endif


static const int	kEndOfLabel = ':';
static char		errorString[80];


/*	Scans through the file pointed to by fp and fills the Token structure
	with the next token found in fp. Return value is the current line number.
	call Scanner with the parameter *inLine = 0 to reset the line count			*/
int Scanner( TokenPtr tok, FILE *fp, int *inLine )
{
	static int	fresh = TRUE;			/* Signals new reaction.				*/
	static int	line = 1;				/* Number of lines processed.			*/
	char		*p;
	int			c, count;
	
#	ifdef applec
	RotateCursor( 32 );
#	endif

	if ( !*inLine ) line = 1;			/* Reset the line count.				*/

	c = getc( fp );
	while ( isspace( c ) ) {			/* Eat up white space chararcters.		*/
		if ( c == '\n' ) {
			++line;
			fresh = TRUE;
		}
		c = getc( fp );
	}
	
	while ( c == '#' ) {				/* Comment line, skip til end of line.	*/
		do {
			c = getc( fp );
			if ( c == EOF ) return FALSE;
		}
		while ( c != '\n' );
		fresh = TRUE;
		++line;							/* Increment number of lines			*/
		do {							/*   and eat the following white space.	*/
			c = getc( fp );
			if ( c == '\n' ) {
				++line;
				fresh = TRUE;
			}
		} while ( isspace( c ) );
	}
	
	if ( c == EOF ) {
		tok->type = kDone;
		*inLine = line;
		return FALSE;					/* End of file found.					*/
	}
	if ( fresh ) {						/* First thing we expect is a label.	*/
		p = tok->string;
		count = 1;						/* Add 1 for terminating '\0'.			*/
		while ( c != kEndOfLabel ) {
			if ( ++count >= kMaxLen ) {
				sprintf( errorString, "too many characters in label!" );
				ParseError( line, errorString );
			}
			*p++ = c;
			c = getc( fp );
			if ( c == '\n' || c == EOF ) {
				sprintf( errorString, "expected a \":\" after the label!" );
				ParseError( line, errorString );
			}
		}
		*p = '\0';						/* Add null to end of string.			*/
		tok->type = kLabel;
		fresh = FALSE;
		*inLine = line;
		return TRUE;

		/*	c now contains the character ':', which can be discarded.		 	*/
	}
	
	if ( isdigit( c ) || c == '.' ) {	/* Found a number.						*/
		ungetc( c, fp );				/* Put the character back.				*/
		fscanf( fp, "%lf", &tok->value );
		tok->type = kNumber;
		*inLine = line;
		return TRUE;
	}
	
	if ( c == '+' ) {					/* Found a '+' sign.					*/
		tok->type = kPlus;
		tok->string[0] = c;
		tok->string[1] = '\0';
		*inLine = line;
		return TRUE;
	}
	
	if ( c == '-' ) {					/* Found a possible reaction direction	*/
		p = tok->string;				/*    or a '-' sign.					*/
		*p++ = c;
		if ( ( c = getc( fp ) ) != '>' ) {	/* It is a '-' sign.				*/
			ungetc( c, fp );				/* Put the character back.			*/
			tok->type = kMinus;
			tok->string[0] = '-';
			tok->string[1] = '\0';
			*inLine = line;
			return TRUE;
		}
		*p++ = c;
		*p = '\0';
		tok->type = kReactionDirection;
		*inLine = line;
		return TRUE;
	}
	
	if ( c == '=' ) {					/* Next poss. reaction direction found.	*/
		p = tok->string;
		*p++ = c;
		c = getc( fp );
		if (  c == '=' || c == '>' ) {
			*p++ = c;
		} else {
			ungetc( c, fp );
		}
		*p = '\0';
		tok->type = kReactionDirection;
		*inLine = line;
		return TRUE;
	}

	if ( isalpha( c ) ) {				/* Found an identifier.					*/
		p = tok->string;
		count = 1;
		do {
			if ( ++count >= kMaxLen ) {
				sprintf( errorString, "too many characters in identifier!" );
				ParseError( line, errorString );
			}
			*p++ = c;
			c = getc( fp );

			if ( c == '-' ) {
				int		d = getc( fp );
				
				if ( isalpha( d ) ) {		/* isomere prefix		*/
					*p++ = c;
					c = d;
				}
				else {						/* reaction direction	*/
					ungetc( d, fp );
				}
			}
		} while ( isalnum(c) && c != EOF );
		ungetc( c, fp );
		*p = '\0';
		tok->type = kIdentifier;
		*inLine = line;
		return TRUE;
	}
	
	/*	If we arrive here, c contains a character which doesn't make sense,
		so issue an error message.												*/
	sprintf( errorString, "unrecognized character \"%c\"", c );
	ParseError( line, errorString );
}

